# vim: ts=4:sts=4:sw=4
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

import typing

__description__ = ('The idf_converter package provides tools to convert Earth '
                   'Observation data into IDF files.')
__version__ = '0.1'
__author__ = 'Lucile Gaultier, Gilles Guitton, Sylvain Herlédan'
__author_email__ = 'idf-support@oceandatalab.com'
__url__ = 'https://seascope.oceandatalab.com'
__keywords__: typing.List[str] = []
__classifiers__ = ['Development Status :: 5 - Production/Stable',
                   'Environment :: Console',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU Lesser General Public '
                   'License v3 or later (LGPLv3+)',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: Implementation :: '
                   'CPython',
                   'Topic :: Scientific/Engineering :: GIS',
                   'Typing :: Typed', ]
